/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice
 * (including the next paragraph) shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL VIA, S3 GRAPHICS, AND/OR
 * ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include "via_chrome9_drv.h"
#include "via_chrome9_object.h"
#include "via_chrome9_drm.h"
static struct vm_operations_struct via_chrome9_ttm_vm_ops;
static const struct vm_operations_struct *ttm_vm_ops = NULL;
extern struct drm_device *drm_dev_v4l;

/**
 * initialize the gem buffer object
 */
int via_chrome9_gem_object_init(struct drm_gem_object *obj)
{
	return 0;
}

/**
 * free the gem buffer object
 */
void via_chrome9_gem_object_free(struct drm_gem_object *gobj)
{
	struct via_chrome9_object *vobj = gobj->driver_private;

	gobj->driver_private = NULL;
	if (vobj) {
		via_chrome9_buffer_object_unref(&vobj);
	}
}

/**
 * Unreference the ttm buffer object and do ttm_bo_release while this bo kref equal to zero
 */
void
via_chrome9_buffer_object_unref(struct via_chrome9_object **vobj)
{
	struct ttm_buffer_object *bo;

	if ((*vobj) == NULL) {
		return;
	}
	bo = &((*vobj)->bo);
	ttm_bo_unref(&bo);

	if (NULL == bo)
		*vobj = NULL;
}
EXPORT_SYMBOL(via_chrome9_buffer_object_unref);

/**
 * Add a reference to this buffer object
 */
void
via_chrome9_buffer_object_ref(struct ttm_buffer_object *bo)
{
	bo = ttm_bo_reference(bo);
}

/**
 * Unmap the ttm buffer object
 */
void
via_chrome9_buffer_object_kunmap(struct via_chrome9_object *vobj)
{
	spin_lock(&vobj->bo.lock);
	if (vobj->ptr == NULL) {
		spin_unlock(&vobj->bo.lock);
		return;
	}
	vobj->ptr = NULL;
	spin_unlock(&vobj->bo.lock);
	ttm_bo_kunmap(&vobj->kmap);
}
EXPORT_SYMBOL(via_chrome9_buffer_object_kunmap);

/**
 * Register to the bo->destroy and destroy the via_chrome9_object while list_kref equal to zero
 */
static void
via_chrome9_buffer_object_destroy(struct ttm_buffer_object *bo)
{
	struct via_chrome9_object *vobj;

	vobj = container_of(bo, struct via_chrome9_object, bo);
	kfree(vobj);
	vobj = NULL;
}

/**
 * Creat a via_chrome9_buffer_object
 */
int 
via_chrome9_buffer_object_create(struct ttm_bo_device *bdev,
			unsigned long size,
			enum ttm_bo_type type,
			uint32_t flags,
			uint32_t page_alignment,
			unsigned long buffer_start,
			bool interruptible,
			struct file *persistant_swap_storage,
			struct via_chrome9_object **p_bo)
{
	struct drm_via_chrome9_private *dev_priv;
	struct via_chrome9_object *vobj;
	int ret;

	dev_priv = container_of(bdev, struct drm_via_chrome9_private, bdev);
	if (unlikely(bdev->dev_mapping == NULL)) {
		bdev->dev_mapping = dev_priv->ddev->dev_mapping;
	}

	vobj = kzalloc(sizeof(struct via_chrome9_object), GFP_KERNEL);
	if (unlikely(vobj == NULL)) {
		return -ENOMEM;
	}

	ret = ttm_buffer_object_init(bdev, &vobj->bo, size, type, flags,
				     page_alignment, buffer_start,
				     interruptible,
				     persistant_swap_storage, size,
				     &via_chrome9_buffer_object_destroy);
	if (unlikely(ret != 0)) {
		DRM_ERROR("Failed to allocate TTM object %ld  \n",size);
		return ret;
	}

	*p_bo = vobj;

	return 0;
}
/*For v4l module*/
int 
via_chrome9_buffer_object_create2(unsigned long size,
			enum ttm_bo_type type,
			uint32_t flags,
			uint32_t page_alignment,
			unsigned long buffer_start,
			bool interruptible,
			struct file *persistant_swap_storage,
			struct via_chrome9_object **p_bo)
{
	struct drm_via_chrome9_private *dev_priv =
		(struct drm_via_chrome9_private *)drm_dev_v4l->dev_private;

	return via_chrome9_buffer_object_create(&dev_priv->bdev, size, type,
			flags, page_alignment, buffer_start, interruptible,
			persistant_swap_storage, p_bo);
}
EXPORT_SYMBOL(via_chrome9_buffer_object_create2);

/**
 * wait this bo idle
 */
int via_chrome9_object_wait(struct via_chrome9_object *vobj, bool no_wait)
{
	int ret = 0;
	ret = ttm_bo_reserve(&vobj->bo, true, false, false, 0);
	if (unlikely(ret != 0)) {
		DRM_ERROR("failed reserve the bo for object wait\n");
		return ret;
	}
	spin_lock(&vobj->bo.lock);
	if (vobj->bo.sync_obj) {
		ret = ttm_bo_wait(&vobj->bo, true, false, no_wait);
	}
	/* clear command flush flag */
	if (0 == ret)
		clear_bit(VIA_CHROME9_BO_FLAG_CMD_FLUSHING, &vobj->flags);

	spin_unlock(&vobj->bo.lock);
	ttm_bo_unreserve(&vobj->bo);

	return ret;
}

/**
  * kmap this buffer object and return the virtual address
  */
int 
via_chrome9_buffer_object_kmap(struct via_chrome9_object *vobj, void **ptr)
{
	int ret;

	spin_lock(&vobj->bo.lock);
	if (vobj->ptr) {
		if (ptr) {
			*ptr = vobj->ptr;
		}
		spin_unlock(&vobj->bo.lock);
		return 0;
	}
	spin_unlock(&vobj->bo.lock);

	ret = ttm_bo_reserve(&vobj->bo, false, false, false, 0);
	if (ret)
		return ret;

	ret = ttm_bo_kmap(&vobj->bo, 0, vobj->bo.num_pages, &vobj->kmap);
	if (ret) {
		DRM_ERROR(" kmap the buffer object error \n");
		ttm_bo_unreserve(&vobj->bo);
		return ret;
	}
	ttm_bo_unreserve(&vobj->bo);

	spin_lock(&vobj->bo.lock);
	vobj->ptr = ttm_kmap_obj_virtual(&vobj->kmap, &vobj->iomem);
	spin_unlock(&vobj->bo.lock);
	if (ptr) {
		*ptr = vobj->ptr;
	}

	return 0;
}

EXPORT_SYMBOL(via_chrome9_buffer_object_kmap);


/**
  * use mmap to map this buffer object
  */
int via_chrome9_object_mmap(struct drm_file *file_priv,
				struct via_chrome9_object *vobj,
				uint64_t size, uint64_t *offset, void **virtual)
{
	*offset = vobj->bo.addr_space_offset;
	*virtual = (void *)do_mmap_pgoff(file_priv->filp, 0, size,
		PROT_READ | PROT_WRITE, MAP_SHARED,
		vobj->bo.addr_space_offset >> PAGE_SHIFT);
	if (*virtual == ((void *)-1))
		return  -ENOMEM;

	return 0;
}

static int via_chrome9_ttm_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
{
	struct ttm_buffer_object *bo;
	int r;

	bo = (struct ttm_buffer_object *)vma->vm_private_data;
	if (bo == NULL)
		return VM_FAULT_NOPAGE;

	r = ttm_vm_ops->fault(vma, vmf);
	return r;
}

/**
  * file operation mmap
  */
int via_chrome9_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct drm_file *file_priv;
	struct drm_via_chrome9_private *dev_priv;
	int ret;

	if (unlikely(vma->vm_pgoff < DRM_FILE_PAGE_OFFSET)) 
		return drm_mmap(filp, vma);

	file_priv = (struct drm_file *)filp->private_data;
	dev_priv = file_priv->minor->dev->dev_private;
	if (dev_priv == NULL)
		return -EINVAL;

	ret = ttm_bo_mmap(filp, vma, &dev_priv->bdev);
	if (unlikely(ret != 0))
		return ret;

	if (unlikely(ttm_vm_ops == NULL)) {
		ttm_vm_ops = vma->vm_ops;
		via_chrome9_ttm_vm_ops = *ttm_vm_ops;
		via_chrome9_ttm_vm_ops.fault = &via_chrome9_ttm_fault;
	}
	vma->vm_ops = &via_chrome9_ttm_vm_ops;

	return 0;
}

/**
 * Allocate the bo for vq gart table and ring buffer
 */
int via_chrome9_allocate_basic_bo(struct drm_device *dev)
{
	struct drm_via_chrome9_private *dev_priv =
		(struct drm_via_chrome9_private *)dev->dev_private;
	int ret;

	/* allocate vq bo */
	ret = via_chrome9_buffer_object_create(&dev_priv->bdev, DEV_VQ_MEMORY,
				       ttm_bo_type_kernel,
				       TTM_PL_FLAG_VRAM |
				       TTM_PL_FLAG_NO_EVICT,
				       0, 0, false, NULL, &dev_priv->vq);
	if (unlikely(ret))
		return ret;

	/* allocate gart table bo */
	ret = via_chrome9_buffer_object_create(&dev_priv->bdev,
					(dev_priv->pcie_mem_size >> 10),
					ttm_bo_type_kernel,
					TTM_PL_FLAG_VRAM |TTM_PL_FLAG_NO_EVICT,
					0, 0, false, NULL, &dev_priv->agp_gart);
	if (unlikely(ret))
		goto out_err0;
	dev_priv->pcie_gart_start = dev_priv->agp_gart->bo.offset;

	/* kmap the gart table */
	ret = via_chrome9_buffer_object_kmap(dev_priv->agp_gart,
						(void **)&dev_priv->pcie_gart_map);
	if (ret) {
		DRM_ERROR("kmap the gart table error");
		goto out_err1;
	}

	/* allocate ring buffer here */
	dev_priv->ring_buffer_size = DEV_COMMAND_BUFFER_MEMORY;
	ret = via_chrome9_buffer_object_create(&dev_priv->bdev,
					dev_priv->ring_buffer_size,
					ttm_bo_type_kernel,
					TTM_PL_FLAG_TT |
					TTM_PL_FLAG_NO_EVICT,
					0, 0, false, NULL, &dev_priv->agp_ringbuffer);
	if (unlikely(ret)) {
		DRM_ERROR("allocate the ringbuffer error");
		goto out_err2;
	}

	/* kmap the ring buffer */
	ret = via_chrome9_buffer_object_kmap(dev_priv->agp_ringbuffer,
					(void **)&dev_priv->pcie_ringbuffer_map);
	if (unlikely(ret)) {
		DRM_ERROR("kmap the ringbuffer error");
		goto out_err3;
	}

	/*allocate shadow gart-table for apci store*/
	ret = via_chrome9_buffer_object_create(&dev_priv->bdev,
					(dev_priv->pcie_mem_size >> 10),
					ttm_bo_type_kernel,
					TTM_PL_FLAG_TT |TTM_PL_FLAG_NO_EVICT,
					0, 0, false, NULL, &dev_priv->pm_backup.agp_gart_shadow);
	if (unlikely(ret))
		goto out_err3;

	ret = via_chrome9_buffer_object_kmap(dev_priv->pm_backup.agp_gart_shadow, NULL);
	if (ret) {
		DRM_ERROR("kmap the gart table error");
		goto out_err4;
	}
	/*FIXME, patch before using KMS*/
	dev_priv->gart_valid = true;
	return 0;

	out_err4:
		via_chrome9_buffer_object_unref(&dev_priv->pm_backup.agp_gart_shadow);
	out_err3:
		via_chrome9_buffer_object_unref(&dev_priv->agp_ringbuffer);
	out_err2:
		via_chrome9_buffer_object_kunmap(dev_priv->agp_gart);
		dev_priv->pcie_gart_map = NULL;
	out_err1:
		via_chrome9_buffer_object_unref(&dev_priv->agp_gart);
	out_err0:
		via_chrome9_buffer_object_unref(&dev_priv->vq);
	return ret;
}

/*evict vram BO (can be evicted) to system memory*/
int via_chrome9_evict_vram(struct drm_device *dev)
{
	struct drm_via_chrome9_private *dev_priv =
		(struct drm_via_chrome9_private *)dev->dev_private;

	return ttm_bo_evict_mm(&dev_priv->bdev, TTM_PL_VRAM);
}


static inline uint32_t via_chrome9_object_flags_from_domain(uint32_t domain)
{
	uint32_t flags = 0;
	if (domain & VIA_CHROME9_GEM_DOMAIN_VRAM) {
		flags |= TTM_PL_FLAG_VRAM | TTM_PL_FLAG_WC | TTM_PL_FLAG_UNCACHED;
	}
	if (domain & VIA_CHROME9_GEM_DOMAIN_GTT) {
		flags |= TTM_PL_FLAG_TT | TTM_PL_FLAG_WC | TTM_PL_FLAG_UNCACHED;
	}
	if (domain & VIA_CHROME9_GEM_DOMAIN_CPU) {
		flags |= TTM_PL_FLAG_SYSTEM | TTM_PL_MASK_CACHING;
	}
	if (!flags) {
		flags |= TTM_PL_FLAG_SYSTEM | TTM_PL_MASK_CACHING;
	}
	return flags;
}
/*Pin a BO to a memory type with NO_EVICT*/
int via_chrome9_bo_pin(struct via_chrome9_object *bo, u32 domain, u64 *gpu_addr)
{
	int r;
	uint32_t flags, tmp;

	r = ttm_bo_reserve(&bo->bo, true, false, false, 0);
	if (unlikely(r!=0))
		return r;

	flags = via_chrome9_object_flags_from_domain(domain);
	tmp = bo->bo.mem.placement;
	ttm_flag_masked(&tmp, flags, TTM_PL_MASK_MEM);
	bo->bo.proposed_placement = tmp | TTM_PL_FLAG_NO_EVICT | TTM_PL_MASK_CACHING;
	r = ttm_buffer_object_validate(&bo->bo, bo->bo.proposed_placement, false, false);
	if (likely(r == 0)) {
		if (gpu_addr != NULL)
			*gpu_addr = bo->bo.offset;
	}
	if (unlikely(r != 0))
		DRM_ERROR( "via_chrome9 pin failed\n");

	ttm_bo_unreserve(&bo->bo);
	return r;
}
EXPORT_SYMBOL(via_chrome9_bo_pin);

/*Remove the NO_EVICT attribute of a BO*/

int via_chrome9_bo_unpin(struct via_chrome9_object *bo)
{
	int r;
	uint32_t flags;

	r = ttm_bo_reserve(&bo->bo, true, false, false, 0);
	if (unlikely(r!=0))
		return r;

	flags = bo->bo.mem.placement;
	bo->bo.proposed_placement = flags & ~TTM_PL_FLAG_NO_EVICT;
	r = ttm_buffer_object_validate(&bo->bo, bo->bo.proposed_placement, false, false);
	if (unlikely(r != 0))
		DRM_ERROR( "via_chrome9 unpin failed\n");
	
	ttm_bo_unreserve(&bo->bo);
	return r;
}
EXPORT_SYMBOL(via_chrome9_bo_unpin);

/*For BO domain changes*/
/*GTT BO(write-combine), unbind to system memory(cached) */
/*can accel CPU access*/
int via_chrome9_object_set_domain(struct via_chrome9_object *bo,
		u32 domain)
{
	int r;
	uint32_t flags, tmp;

	r = ttm_bo_reserve(&bo->bo, true, false, false, 0);
	if (unlikely(r!=0))
		return r;

	flags = via_chrome9_object_flags_from_domain(domain);
	tmp = bo->bo.mem.placement;
	ttm_flag_masked(&tmp, flags, TTM_PL_MASK_MEM);
	bo->bo.proposed_placement = tmp | TTM_PL_MASK_CACHING;
	r = ttm_buffer_object_validate(&bo->bo, bo->bo.proposed_placement, false, false);

	if (unlikely(r != 0))
		DRM_ERROR( "via_chrome9 set domain= %x failed\n",domain);
	ttm_bo_unreserve(&bo->bo);
	return r;
}
/*GPU wait CPU*/
int via_chrome9_object_wait_cpu_access(struct via_chrome9_object *bo,
		struct drm_file *file_priv)
{
	int ret = 0;

	if(unlikely((atomic_read(&bo->bo.cpu_writers) > 0))){
		if (bo->owner_file){
			if (file_priv != bo->owner_file)
				ret = ttm_bo_wait_cpu(&bo->bo, false);
			else
				BUG_ON(file_priv == bo->owner_file);
		}
	}
	return ret;
}
/*since kernel-2.6.32 do not export those function to drivers, we add it by ourself*/
int ttm_bo_wait_cpu(struct ttm_buffer_object *bo, bool no_wait)
{
	if ((atomic_read(&bo->cpu_writers) > 0) && no_wait)
		return -EBUSY;

	return wait_event_interruptible(bo->event_queue,
					atomic_read(&bo->cpu_writers) == 0);
}
int ttm_bo_synccpu_write_grab(struct ttm_buffer_object *bo, bool no_wait)
{
	int ret = 0;

	/*
	 * Using ttm_bo_reserve instead of ttm_bo_block_reservation
	 * makes sure the lru lists are updated.
	 */

	ret = ttm_bo_reserve(bo, true, no_wait, false, 0);
	if (unlikely(ret != 0))
		return ret;
	spin_lock(&bo->lock);
	ret = ttm_bo_wait(bo, false, true, no_wait);
	spin_unlock(&bo->lock);
	if (likely(ret == 0))
		atomic_inc(&bo->cpu_writers);
	ttm_bo_unreserve(bo);
	return ret;
}
void ttm_bo_synccpu_write_release(struct ttm_buffer_object *bo)
{
	if (atomic_dec_and_test(&bo->cpu_writers))
		wake_up_all(&bo->event_queue);
}
